from .base_enums import AvatarPart


class ClothingGraphic(AvatarPart):
    """Clothing graphics. The graphics printed on the clothing if the selected
clothing type is ClothingType.GRAPHIC_SHIRT"""

    __install__ = True
    __enum_path__ = 'clothing_graphics.py'
    __path__ = 'avatar_parts/clothes/graphic'

    NONE = ''
    BAT = 'bat'
    BEAR = 'bear'
    CUMBIA = 'cumbia'
    CUSTOM_TEXT = 'custom_text'
    DEER = 'deer'
    DIAMOND = 'diamond'
    HOLA = 'hola'
    PIZZA = 'pizza'
    RESIST = 'resist'
    SELENA = 'selena'
    SKULL_OUTLINE = 'skull_outline'
    SKULL = 'skull'



from .base_enums import AvatarColor


class ClothingColor(AvatarColor):
    """Clothing colors. You can use any other color using its hex code"""

    __install__ = True
    __enum_path__ = 'clothing_colors.py'

    BLACK = '#262E33'
    BLUE_01 = '#65C9FF'
    BLUE_02 = '#5199E4'
    BLUE_03 = '#25557C'
    GRAY_01 = '#E6E6E6'
    GRAY_02 = '#929598'
    HEATHER = '#3C4F5C'
    PASTEL_BLUE = '#B1E2FF'
    PASTEL_GREEN = '#A7FFC4'
    PASTEL_ORANGE = '#FFDEB5'
    PASTEL_YELLOW = '#FFFFB1'
    PINK = '#FF488E'
    RED = '#FF5C5C'
    WHITE = '#FFFFFF'



from .base_enums import AvatarPart


class AvatarStyle(AvatarPart):
    """Avatar styles"""

    __install__ = True
    __enum_path__ = 'avatar_styles.py'
    __path__ = 'avatar_parts/styles'

    TRANSPARENT = 'transparent'
    CIRCLE = 'circle'



from .base_enums import AvatarPart
from .hair_types import HairType
from .hat_types import HatType


class TopType(AvatarPart):
    """
    Hair/top of head types
    """
    __install__ = False
    __enum_path__ = 'top_types.py'
    __path__ = 'avatar_parts/top'

    @staticmethod
    def get_all():
        return list(HairType) + list(HatType)



import os
import random
import pytest

import python_avatars as pa


def test_avatar_generation():
    assert pa.Avatar().render("avatar.svg") != ""
    assert os.path.isfile("avatar.svg")


def test_random_avatar():
    random.seed(123)

    avatar_1 = pa.Avatar.random()
    avatar_2 = pa.Avatar.random()

    assert avatar_1 != avatar_2


def test_not_installable():
    with pytest.raises(RuntimeError):
        pa.install_part("examples/install/suit.svg", pa.TopType)


def test_install_installed_part():
    pa.install_part("examples/install/suit.svg", pa.ClothingType)
    with pytest.raises(FileExistsError):
        pa.install_part("examples/install/suit.svg", pa.ClothingType)


def test_cross_install():
    with pytest.raises(RuntimeError):
        pa.install_part("examples/install/suit.svg", pa.HairColor)


def test_bad_install():
    with pytest.raises(FileNotFoundError):
        pa.install_part("this_file_does_not_exist.svg", pa.HairType)


def test_invalid_color():
    with pytest.raises(RuntimeError):
        pa.install_color("INVALID_1", "#FF00", pa.ClothingColor)

    with pytest.raises(RuntimeError):
        pa.install_color("INVALID_2", "#fa08g5", pa.ClothingColor)

    with pytest.raises(RuntimeError):
        pa.install_color("INVALID_WHITE", "ffffff", pa.ClothingColor)



"""
SVG avatar generation library

More info @ https://github.com/ibonn/python_avatars
"""

__version__ = "1.3.0"

# Enums
from .accessory_types import AccessoryType
from .avatar_styles import AvatarStyle
from .background_colors import BackgroundColor
from .clothing_colors import ClothingColor
from .clothing_graphics import ClothingGraphic
from .clothing_types import ClothingType
from .eye_types import EyeType
from .eyebrow_types import EyebrowType
from .facial_hair_types import FacialHairType
from .hair_colors import HairColor
from .hair_types import HairType
from .hat_types import HatType
from .mouth_types import MouthType
from .nose_types import NoseType
from .skin_colors import SkinColor
from .top_types import TopType

# Classes
from .avatar import Avatar

# Core functions
from .core import install_part
from .core import install_color
from .core import uninstall_part
from .core import uninstall_color
from .core import factory_reset



from .base_enums import AvatarPart


class EyeType(AvatarPart):
    """Eye types"""

    __install__ = True
    __enum_path__ = 'eye_types.py'
    __path__ = 'avatar_parts/eyes'

    CLOSED = 'closed'
    CRY = 'cry'
    DEFAULT = 'default'
    EYE_ROLL = 'eye_roll'
    HAPPY = 'happy'
    HEART = 'heart'
    SIDE = 'side'
    SQUINT = 'squint'
    SURPRISED = 'surprised'
    WINK_WACKY = 'wink_wacky'
    WINK = 'wink'
    X_DIZZY = 'x_dizzy'



from .base_enums import AvatarColor


class HairColor(AvatarColor):
    """Hair/facial hair colors. You can use any other color using its hex code"""

    __install__ = True
    __enum_path__ = 'hair_colors.py'

    AUBURN = '#A55728'
    BLACK = '#2C1B18'
    BLONDE = '#B58143'
    BLONDE_GOLDEN = '#D6B370'
    BROWN = '#724133'
    BROWN_DARK = '#4A312C'
    PASTEL_PINK = '#F59797'
    PLATINUM = '#ECDCBF'
    RED = '#C93305'
    SILVER_GRAY = '#E8E1E1'



from .base_enums import AvatarPart


class FacialHairType(AvatarPart):
    """Facial hair types"""

    __install__ = True
    __enum_path__ = 'facial_hair_types.py'
    __path__ = 'avatar_parts/facial_hair'

    NONE = ''
    BEARD_LIGHT = 'beard_light'
    BEARD_MAGESTIC = 'beard_magestic'
    BEARD_MEDIUM = 'beard_medium'
    MOUSTACHE_FANCY = 'moustache_fancy'
    MOUSTACHE_MAGNUM = 'moustache_magnum'



import enum

from .accessory_types import AccessoryType
from .avatar_styles import AvatarStyle
from .background_colors import BackgroundColor
from .clothing_colors import ClothingColor
from .clothing_graphics import ClothingGraphic
from .clothing_types import ClothingType
from .eye_types import EyeType
from .eyebrow_types import EyebrowType
from .facial_hair_types import FacialHairType
from .hair_colors import HairColor
from .hair_types import HairType
from .hat_types import HatType
from .mouth_types import MouthType
from .nose_types import NoseType
from .skin_colors import SkinColor
from .top_types import TopType

from . svg_parser import SVGParser
from . core import _get_path


class Avatar:
    """
    Avatar. Create a new avatar using this class.
    """

    def __init__(
        self,
        style=AvatarStyle.TRANSPARENT,
        background_color=BackgroundColor.DEFAULT,
        top=HairType.SHORT_FLAT,
        hat_color=ClothingColor.HEATHER,
        eyebrows=EyebrowType.DEFAULT,
        eyes=EyeType.DEFAULT,
        nose=NoseType.DEFAULT,
        mouth=MouthType.DEFAULT,
        facial_hair=FacialHairType.NONE,
        skin_color=SkinColor.LIGHT,
        hair_color=HairColor.BROWN,
        facial_hair_color=HairColor.BROWN,
        accessory=AccessoryType.NONE,
        clothing=ClothingType.pick_random(),
        clothing_color=ClothingColor.HEATHER,
        shirt_graphic=ClothingGraphic.SKULL,
        shirt_text='Hola!',
        title=None,
    ):
        self.style = style
        self.background_color = background_color

        self.top = top
        self.hat_color = hat_color
        self.eyebrows = eyebrows
        self.eyes = eyes
        self.nose = nose
        self.mouth = mouth
        self.facial_hair = facial_hair

        self.skin_color = skin_color
        self.hair_color = hair_color
        self.facial_hair_color = facial_hair_color

        self.accessory = accessory
        self.clothing = clothing
        self.clothing_color = clothing_color
        self.shirt_graphic = shirt_graphic
        self.shirt_text = shirt_text

        self.title = title

    @staticmethod
    def random(
        style=None,
        background_color=None,
        top=None,
        hat_color=None,
        eyebrows=None,
        eyes=None,
        nose=None,
        mouth=None,
        facial_hair=None,
        skin_color=None,
        hair_color=None,
        facial_hair_color=None,
        accessory=None,
        clothing=None,
        clothing_color=None,
        shirt_graphic=None,
        shirt_text=None,
        title=None,
    ):
        """
        Generates a random avatar. The parameters for this method are exactly the same
        as for the constructor. If none of them are specified, the values are chosen
        randomly. The values for the specified parameters will stay fixed
        """
        return Avatar(
            style=AvatarStyle.pick_random() if style is None else style,
            background_color=BackgroundColor.pick_random(
            ) if background_color is None else background_color,
            top=TopType.pick_random() if top is None else top,
            hat_color=ClothingColor.pick_random() if hat_color is None else hat_color,
            eyebrows=EyebrowType.pick_random() if eyebrows is None else eyebrows,
            eyes=EyeType.pick_random() if eyes is None else eyes,
            nose=NoseType.pick_random() if nose is None else nose,
            mouth=MouthType.pick_random() if mouth is None else mouth,
            facial_hair=FacialHairType.pick_random(
                favor=FacialHairType.NONE) if facial_hair is None else facial_hair,

            skin_color=SkinColor.pick_random() if skin_color is None else skin_color,
            hair_color=HairColor.pick_random() if hair_color is None else hair_color,
            facial_hair_color=HairColor.pick_random(
            ) if facial_hair_color is None else facial_hair_color,


            accessory=AccessoryType.pick_random(
                favor=AccessoryType.NONE) if accessory is None else accessory,
            clothing=ClothingType.pick_random() if clothing is None else clothing,
            clothing_color=ClothingColor.pick_random(
            ) if clothing_color is None else clothing_color,
            shirt_graphic=ClothingGraphic.pick_random(
            ) if shirt_graphic is None else shirt_graphic,
            shirt_text=shirt_text,

            title=title,
        )

    def render(self, path=None):
        """
        Render the avatar to svg. Always returns the resulting svg as a string

        :param path: The path where the svg file will be saved. If ``None``, no file is saved
        :type path: str
        :return: str
        """

        # Load the base template based on the avatar style
        avatar = SVGParser(_get_path(AvatarStyle, self.style))

        if self.title is not None:
            avatar.get_element_by_id("Title").set_content(self.title)

        # Set style params
        bg_color = avatar.get_element_by_id("Background-Color")
        if bg_color is not None:
            bg_color.set_attr("fill", self.background_color)

        # Set skin color
        avatar.get_element_by_id(
            'Skin-Color').set_attr('fill', self.skin_color)

        # Set top
        if not self.__is_empty(self.top):

            if isinstance(self.top, HairType):
                top = SVGParser(_get_path(HairType, self.top))

                hair_color = top.get_element_by_id(
                    'Hair-Color'
                )

                if hair_color is not None:
                    hair_color.set_attr('fill', self.hair_color)

            else:
                top = SVGParser(_get_path(HatType, self.top))
                hat_color = top.get_element_by_id("Fabric-Color")
                if hat_color is not None:
                    hat_color.set_attr('fill', self.hat_color)

            # Set facial hair (top)
            if not self.__is_empty(self.facial_hair):
                top_facial_hair = top.get_element_by_id('Facial-Hair')
                if top_facial_hair:
                    facial_hair = SVGParser(
                        _get_path(FacialHairType, self.facial_hair)
                    )

                    facial_hair.get_element_by_id(
                        'Facial-Hair-Color'
                    ).set_attr('fill', self.facial_hair_color)

                    top_facial_hair.set_content(facial_hair.children())

            # Set accessories (top)
            if not self.__is_empty(self.accessory):
                accessories = SVGParser(
                    _get_path(AccessoryType, self.accessory))
                top.get_element_by_id('Accessory').set_content(
                    accessories.children())

            avatar.get_element_by_id('Top').set_content(top.children())

        # Set eyebrows
        if not self.__is_empty(self.eyebrows):
            eyebrow = SVGParser(_get_path(EyebrowType, self.eyebrows))
            avatar.get_element_by_id('Eyebrow').set_content(eyebrow.children())

        # Set eyes
        if not self.__is_empty(self.eyes):
            eyes = SVGParser(_get_path(EyeType, self.eyes))
            avatar.get_element_by_id('Eyes').set_content(eyes.children())

        # Set nose
        if not self.__is_empty(self.nose):
            nose = SVGParser(_get_path(NoseType, self.nose))
            avatar.get_element_by_id('Nose').set_content(nose.children())

        # Set mouth
        if not self.__is_empty(self.mouth):
            mouth = SVGParser(_get_path(MouthType, self.mouth))
            avatar.get_element_by_id('Mouth').set_content(mouth.children())

        # Set clothes
        if not self.__is_empty(self.clothing):
            clothes = SVGParser(_get_path(ClothingType, self.clothing))
            fabric_color = clothes.get_element_by_id('Fabric-Color')
            if fabric_color:
                fabric_color.set_attr('fill', self.clothing_color)

            if self.clothing == ClothingType.GRAPHIC_SHIRT and not self.__is_empty(self.shirt_graphic):
                graphic = SVGParser(
                    _get_path(ClothingGraphic, self.shirt_graphic)
                )

                if self.shirt_graphic == ClothingGraphic.CUSTOM_TEXT:
                    graphic.get_element_by_id(
                        'Graphic-Text'
                    ).children('tspan')[0].set_content(self.shirt_text)

                clothes.get_element_by_id(
                    'Graphic'
                ).set_content(graphic.children())

            avatar.get_element_by_id(
                'Clothing'
            ).set_content(clothes.children())

        return avatar.render(path)

    def happy(self):
        pass

    def sad(self):
        pass

    def frightened(self):
        pass

    def disgusted(self):
        pass

    def angry(self):
        pass

    def surprised(self):
        pass

    def confused(self):
        pass

    def shocked(self):
        return self.surprised()

    def worried(self):
        pass

    @staticmethod
    def __is_empty(value):
        """
        Check wether a value is set or not for an attribute

        :param value: The value to test
        :type value: str or :class:`AvatarEnum`
        :return: bool
        """
        return value is None or value == '' or isinstance(value, enum.Enum) and value.value == ''

    def __hash__(self):
        return hash(
            (
                self.style,
                self.background_color,
                self.top,
                self.hat_color,
                self.eyebrows,
                self.eyes,
                self.nose,
                self.mouth,
                self.facial_hair,
                self.skin_color,
                self.hair_color,
                self.facial_hair_color,
                self.accessory,
                self.clothing,
                self.clothing_color,
                self.shirt_graphic,
                self.shirt_text,
                self.title,
            )
        )

    def __eq__(self, other):
        return hash(self) == hash(other)

    def __ne__(self, other):
        return not self.__eq__(other)

    def __str__(self):
        """
        String representation for the avatar
        """
        return str(self.__dict__)



from .base_enums import AvatarPart


class HairType(AvatarPart):
    """Hair types"""

    __install__ = True
    __enum_path__ = 'hair_types.py'
    __path__ = 'avatar_parts/top/hair'

    NONE = 'no_hair'
    BIG_HAIR = 'big_hair'
    BOB = 'bob'
    BUN = 'bun'
    CAESAR_SIDE_PART = 'caesar_side_part'
    CAESAR = 'caesar'
    CURLY = 'curly'
    CURVY = 'curvy'
    DREADS = 'dreads'
    FRIDA = 'frida'
    FRIZZLE = 'frizzle'
    FRO_BAND = 'fro_band'
    FRO = 'fro'
    LONG_NOT_TOO_LONG = 'long_not_too_long'
    MIA_WALLACE = 'mia_wallace'
    SHAGGY_MULLET = 'shaggy_mullet'
    SHAGGY = 'shaggy'
    SHAVED_SIDES = 'shaved_sides'
    SHORT_CURLY = 'short_curly'
    SHORT_DREADS_1 = 'short_dreads_1'
    SHORT_DREADS_2 = 'short_dreads_2'
    SHORT_FLAT = 'short_flat'
    SHORT_ROUND = 'short_round'
    SHORT_WAVED = 'short_waved'
    SIDES = 'sides'
    STRAIGHT_1 = 'straight_1'
    STRAIGHT_2 = 'straight_2'
    STRAIGHT_STRAND = 'straight_strand'



import enum
import random


class AvatarEnum(enum.Enum):
    """
    Base enum for the library. Allows picking random elements from the avatar
    """
    __path__ = ''
    __install__ = False
    __enum_path__ = 'base_enums.py'

    @classmethod
    def get_all(cls):
        """
        Get a list containing all the values for this enum
        """
        return list(cls)

    @classmethod
    def pick_random(cls, favor=None):
        """
        Pich a random value from this enum. The value in ``favor`` has
        more chances of being chosen
        """
        if favor is None:
            return cls.__pick_random()
        else:
            r = random.uniform(0, 1)
            if r > 0.5:
                return cls.__pick_random()
            else:
                return favor

    @classmethod
    def __pick_random(cls):
        return random.choice(cls.get_all())

    def __str__(self):
        return self.value


class AvatarPart(AvatarEnum):
    """
    Base enum for avatar parts
    """
    pass


class AvatarColor(AvatarEnum):
    """
    Base enum for avatar colors
    """
    pass



import os
import re
from xml.etree import ElementTree as ET

SVG_NAMESPACE_URI = 'http://www.w3.org/2000/svg'
XLINK_NAMESPACE_URI = 'http://www.w3.org/1999/xlink'

# Register namespaces
ET.register_namespace('', SVG_NAMESPACE_URI)
ET.register_namespace('xlink', XLINK_NAMESPACE_URI)


class SVGParser:

    def __init__(self, path=None, svg=None):
        if path is None and svg is None:
            raise Exception('ArgumentMissingException')

        elif path is None:
            self.tree = svg

        elif svg is None:
            tree = ET.parse(path)
            self.tree = tree.getroot()

            # Rename IDs
            self.prefix = os.path.splitext(os.path.basename(path))[0]
            self.__rename_ids()

    def __rename_ids(self):

        xlink_href = '{{{}}}href'.format(XLINK_NAMESPACE_URI)

        attrs = ['clip-path', 'filter', 'mask']

        for x in self.tree.iter():
            if x.get('id') is not None:
                new_id = '{}_{}'.format(self.prefix, x.get('id'))
                x.set('id', new_id)

            for attr in attrs:
                if x.get(attr) is not None:
                    match = re.match('url\(#(.+)\)', x.get(attr))
                    new_id = 'url(#{}_{})'.format(self.prefix, match.group(1))
                    x.set(attr, new_id)

            if x.get(xlink_href) is not None:
                match = re.match('#(.+)', x.get(xlink_href))
                new_id = '#{}_{}'.format(self.prefix, match.group(1))
                x.set(xlink_href, new_id)

    def get_element_by_id(self, element_id):
        result = self.tree.findall(
            '''.//*[@id='{}_{}']'''.format(self.prefix, element_id))

        if len(result) == 0:
            return None

        return SVGParser(svg=result[0])

    def set_attr(self, name, value):
        self.tree.set(name, str(value))

    def children(self, tag_name=None):

        if tag_name is None:
            return [SVGParser(svg=x) for x in self.tree]

        else:
            result = self.tree.findall(
                '{{{}}}{}'.format(SVG_NAMESPACE_URI, tag_name))

            if len(result) == 0:
                raise Exception('TagNotFoundException: {}'.format(tag_name))

            return [SVGParser(svg=x) for x in result]

    def render(self, path=None):

        svg_str = str(self)

        if path is not None:
            with open(path, 'w', encoding='utf-8') as f:
                f.write(svg_str)

        return svg_str

    def set_content(self, content):
        if content is None:
            pass
        elif isinstance(content, list):
            for elem in content:
                self.set_content(elem)
        elif isinstance(content, str):
            self.tree.text = content
        else:
            self.tree.append(content.tree)

    def __str__(self):
        return ET.tostring(self.tree, encoding='utf-8', method='xml').decode('utf-8')


if __name__ == '__main__':
    parser = SVGParser(path='avatar_parts/styles/avataaar_circle.svg')
    hair = SVGParser(path='avatar_parts/top/bun.svg')
    hair.get_element_by_id(
        'Hair-Color').children('path')[0].set_attr('fill', '#FF0000')
    parser.get_element_by_id('Top').set_content(hair.children())
    parser.render('test.svg')



from .base_enums import AvatarColor


class BackgroundColor(AvatarColor):
    """Background colors. You can use any other color using its hex code"""

    __install__ = True
    __enum_path__ = 'background_colors.py'

    DEFAULT = '#65C9FF'
    BLACK = '#262E33'
    WHITE = '#FFFFFF'



from .base_enums import AvatarPart


class MouthType(AvatarPart):
    """Mouth types"""

    __install__ = True
    __enum_path__ = 'mouth_types.py'
    __path__ = 'avatar_parts/mouth'

    CONCERNED = 'concerned'
    DEFAULT = 'default'
    DISBELIEF = 'disbelief'
    EATING = 'eating'
    GRIMACE = 'grimace'
    SAD = 'sad'
    SCREAM_OPEN = 'scream_open'
    SERIOUS = 'serious'
    SMILE = 'smile'
    TONGUE = 'tongue'
    TWINKLE = 'twinkle'
    VOMIT = 'vomit'



from .base_enums import AvatarPart


class EyebrowType(AvatarPart):
    """Eyebrow types"""

    __install__ = True
    __enum_path__ = 'eyebrow_types.py'
    __path__ = 'avatar_parts/eyebrows'

    NONE = ''
    ANGRY_NATURAL = 'angry_natural'
    ANGRY = 'angry'
    DEFAULT_NATURAL = 'default_natural'
    DEFAULT = 'default'
    FLAT_NATURAL = 'flat_natural'
    FROWN_NATURAL = 'frown_natural'
    RAISED_EXCITED_NATURAL = 'raised_excited_natural'
    RAISED_EXCITED = 'raised_excited'
    SAD_CONCERNED_NATURAL = 'sad_concerned_natural'
    SAD_CONCERNED = 'sad_concerned'
    UNIBROW_NATURAL = 'unibrow_natural'
    UP_DOWN_NATURAL = 'up_down_natural'
    UP_DOWN = 'up_down'



from .base_enums import AvatarColor


class SkinColor(AvatarColor):
    """Skin colors. You can use any other color using its hex code"""

    __install__ = True
    __enum_path__ = 'skin_colors.py'

    TANNED = '#FD9841'
    YELLOW = '#F8D25C'
    PALE = '#FFDBB4'
    LIGHT = '#EDB98A'
    BROWN = '#D08B5B'
    DARK_BROWN = '#AE5D29'
    BLACK = '#614335'



import os
import re
import shutil
import json

from .accessory_types import AccessoryType
from .avatar_styles import AvatarStyle
from .background_colors import BackgroundColor
from .clothing_colors import ClothingColor
from .clothing_graphics import ClothingGraphic
from .clothing_types import ClothingType
from .eye_types import EyeType
from .eyebrow_types import EyebrowType
from .facial_hair_types import FacialHairType
from .hair_colors import HairColor
from .hair_types import HairType
from .hat_types import HatType
from .mouth_types import MouthType
from .nose_types import NoseType
from .skin_colors import SkinColor
from .base_enums import AvatarPart, AvatarColor

# Variables for internal usage. Paths to files/folders
_package_path = os.path.dirname(__file__)
_default_path = os.path.join(_package_path, 'default.json')
_installed_path = os.path.join(_package_path, 'installed.json')


def _check(condition, message, exception=RuntimeError):
    """
    Check wether a condition holds or not. If the condition is false, raises a RuntimeError with the provided message

    :param condition: The condition to test
    :param message: The message to show if the condition is false

    :type condition: bool
    :type message: str

    :raises RuntimeError: The error is raised if the condition in ``condition`` is false
    """
    if condition:
        raise exception(message)


def _install_enum(name, value, part_type, enum_type):
    """
    Adds a value to a enum. Returns True on success. In case if error, an exception is raised

    :param name: The name of the value to add
    :param value: The value to add
    :param part_type: The enum to extend
    :param enum_type: The type of the enum to extend

    :type name: str
    :type value: str
    :type part_type: :class:`AvatarEnum`
    :type enum_type: :class:`Avatarpart` or :class:`AvatarColor`

    :raises RuntimeError: The error is raised if the enum cannot be expanded
    :raises FileExistsError: Raised when the enum contains an item named ``name``

    :return: bool
    """
    _check(not part_type.__install__,
           "{} cannot be expanded".format(part_type.__name__))

    const_name = _sanitize(name).upper()

    # Load defaults
    with open(_default_path, 'r') as f:
        default_values = json.load(f)

    # Load installed
    if not os.path.isfile(_installed_path):
        with open(_installed_path, 'w') as f:
            json.dump({}, f)

    with open(_installed_path, 'r') as f:
        installed_values = json.load(f)

    # Combine installed with default
    if part_type.__name__ in installed_values:
        combined = {**default_values[part_type.__name__],
                    **installed_values[part_type.__name__]}
    else:
        combined = default_values[part_type.__name__]
        installed_values[part_type.__name__] = {}

    # Check wether the value exists or not
    _check(const_name in combined,
           "{}.{} already exists".format(part_type.__name__, const_name), FileExistsError)

    # Add to installed
    installed_values[part_type.__name__][const_name] = value

    # Save installed
    with open(_installed_path, 'w') as f:
        json.dump(installed_values, f, indent=4)

    # Overwrite the enum file
    combined = {**default_values[part_type.__name__],
                **installed_values[part_type.__name__]}
    _write_enum(part_type, combined, enum_type)

    return True


def _prompt_uninstall_confirmation(class_name, enum_name):
    """
    Prompt a deletion confirmation message. Returns ``True`` when the user accepted and ``False`` when declined

    :param class_name: The name of the enum from which the value will be deleted
    :param enum_name: The name of the value to delete

    :type class_name: str
    :type enum_name: str

    :return: bool
    """
    return _prompt_confirmation('Do you really want to uninstall {}.{}?'.format(class_name, enum_name))


def _prompt_confirmation(msg):
    """
    Prompt a confirmation message

    :param msg: The message to show
    :type msg: str

    :return: bool
    """
    confirm_resp = input('{} y/n: '.format(msg))
    return confirm_resp.lower().strip() == 'y'


def _uninstall_enum(part, enum_type):
    """
    Removes a value for a given enum. Returns ``True`` on success

    :param part: The name of the value to remove
    :param enum_type: The type of the enum from which the value will be removed

    :type part: :class:`AvatarEnum`
    :type enum_type: :class:`AvatarPart` or :class:`AvatarColor`

    :return: bool
    """
    if os.path.isfile(_installed_path):
        # Load installed
        with open(_installed_path, 'r') as f:
            installed_values = json.load(f)

        # Load default
        with open(_default_path, 'r') as f:
            default_values = json.load(f)

        # Delete from installed
        del installed_values[part.__class__.__name__][part.name]

        # Combine installed + default
        combined = {**default_values[part.__class__.__name__],
                    **installed_values[part.__class__.__name__]}

        # Update installed
        with open(_installed_path, 'w') as f:
            json.dump(installed_values, f)

        # Rewrite enum
        _write_enum(part.__class__, combined, enum_type)
        return True
    return False


def install_part(part_path, part_type):
    """
    Installs a part (clothing, hair type, eye type...)

    :param part_path: The path to the svg to install
    :param part_type: The enum where the new value will be installed

    :type part_path: str
    :type part_type: :class:`AvatarPart`

    :raise RuntimeError: Whenever the part_type does not have an installation path (in other words, when it is not an AvatarPart or when it is not installable)
    :raise FileNotFoundError: If the file at ``part_path`` does not exist

    :return: bool
    """
    _check(part_type.__path__ == '', "Installation path not found")
    _check(not os.path.isfile(part_path),
           "{} does not exist".format(part_path), FileNotFoundError)
    file_name = _sanitize(os.path.splitext(os.path.basename(part_path))[0])

    new_enum = _install_enum(file_name, file_name, part_type, AvatarPart)

    if new_enum:
        # Copy the file on success
        destination = _get_path(part_type, file_name)
        shutil.copy(part_path, destination)

    return True


def uninstall_part(part, confirm=True):
    """
    Removes an installed part. Returns ``True`` on success, ``False`` otherwise

    :param part: The part to uninstall
    :param confirm: Does this action require confirmation? If ``True``, the user will be asked to confirm the uninstall

    :type part: :class:`AvatarPart`
    :type confirm: bool

    :return: bool
    """
    if _prompt_uninstall_confirmation(part.__class__.__name__, part.name) and _uninstall_enum(part, AvatarPart):
        os.remove(_get_path(part.__class__, part.value))
        return True
    return False


def install_color(name, value, part_type):
    """
    Installs a color. Returns ``True`` on success

    :param name: The name of the color to install
    :param value: The value of the color, hex encoded (i.e. #AF541B)
    :param part_type: The enum where the color will be installed

    :type name: str
    :type value: str
    :type part_type: :class:`AvatarColor`

    :raises RuntimeError: If ``part_type`` is not an :class:`AvatarColor`

    :return: bool

    """
    _check(part_type.__path__ != '',
           "{} is not a color".format(part_type.__name__))

    _check(not re.match(r"^#(?:[a-f0-9]{6}|[a-f0-9]{3})$", value, re.MULTILINE | re.IGNORECASE),
           "{} is not a valid color".format(value))

    return _install_enum(name, value, part_type, AvatarColor)


def uninstall_color(color, confirm=True):
    """
    Unistalls a color. Returns ``True`` on success, ``False`` otherwise

    :param color: The color to remove
    :param confirm: Does this action require confirmation? If ``True``, the user will be asked to confirm the uninstall

    :type color: :class:`AvatarColor`
    :type confirm: bool

    :return: bool

    """
    if confirm:
        return _prompt_uninstall_confirmation(color.__class__.__name__, color.name) and _uninstall_enum(color, AvatarColor)
    return _uninstall_enum(color, AvatarColor)


def factory_reset(confirm=True):
    """
    Resets the package to its original state

    :param confirm: Does this action require confirmation? If ``True``, the user will be asked to confirm the reset
    :type confirm: bool
    """

    if confirm and _prompt_confirmation('Do you want to uninstall all the installed parts?'):
        # Load installed and remove svgs
        if os.path.isfile(_installed_path):
            with open(_installed_path, 'r') as f:
                installed_values = json.load(f)

            for enum_name, enum_values_dict in installed_values.items():
                enum = globals()[enum_name]
                for value in enum_values_dict.values():
                    svg_path = _get_path(enum, value)
                    os.remove(svg_path)

            os.remove(_installed_path)

        # Load defaults
        with open(_default_path, 'r') as f:
            default_values = json.load(f)

        for enum_name, enum_values_dict in default_values.items():
            enum_cls = type(enum_name, (object,), enum_values_dict)
            _write_enum(enum_cls, enum_values_dict,
                        AvatarPart if "__path__" in enum_values_dict else AvatarColor)


def _get_path(enum_cls, value):
    """
    Returns the path to the svg file for an AvatarPart given the enum and the name

    :param enum_cls: The enum
    :param value: The name

    :return: str
    """
    p = os.path.join(_package_path, enum_cls.__path__, '{}.svg'.format(value))
    return p


def _sanitize(value):
    """
    Prepares a string to be used as a Python identifier

    :param value: The string to convert to a valid Python identifier

    :type value: str

    :return: str
    """
    # Remove invalid Python identifier chars
    value = re.sub('[^A-Za-z0-9_]+', '_', value)

    # Remove digits from the beginning
    value = value.lstrip('0123456789')

    return value


def _write_enum(e, values_dict, t):
    """
    Writes an enum to an importable file

    :param e: The original enum to overwrite
    :param values_dict: Dictionary holding the values for the new enum
    :param t: Type of the enum

    :type e: :class:`AvatarEnum`
    :type values_dict: dict
    :type t: :class:`AvatarPart` or :class:`AvatarColor`
    """
    filename = e.__enum_path__
    name = e.__name__
    path = e.__path__ if hasattr(e, "__path__") else None
    install = e.__install__
    docstring = e.__doc__

    file_path = os.path.join(_package_path, filename)

    with open(file_path, 'w') as f:
        if t is AvatarColor:
            f.write('from .base_enums import AvatarColor\n\n\n')
            f.write('class {}(AvatarColor):\n'.format(name))
        elif t is AvatarPart:
            f.write('from .base_enums import AvatarPart\n\n\n')
            f.write('class {}(AvatarPart):\n'.format(name))
        else:
            raise RuntimeError()

        f.write('    """{}"""\n\n'.format(docstring))

        f.write("    __install__ = {}\n".format(install))
        f.write("    __enum_path__ = '{}'\n".format(filename))
        if path is not None:
            f.write("    __path__ = '{}'\n".format(path))
        f.write("\n")
        for name, value in values_dict.items():
            if name[:2] != '__':
                f.write("    {} = '{}'\n".format(name, value))



from .base_enums import AvatarPart


class NoseType(AvatarPart):
    """Nose types"""

    __install__ = True
    __enum_path__ = 'nose_types.py'
    __path__ = 'avatar_parts/nose'

    DEFAULT = 'default'



from setuptools import setup

from python_avatars import __version__

with open('README.md') as readme_file:
    README = readme_file.read()

setup_args = dict(
    name='python_avatars',
    version=__version__,
    description='SVG avatar library for Python',
    long_description_content_type="text/markdown",
    long_description=README,
    license='MIT',
    packages=['python_avatars'],
    package_dir={
        'python_avatars': 'python_avatars'
    },
    package_data={'python_avatars': [
        '*.json',
        'avatar_parts/**/*.svg',
        'avatar_parts/**/**/*.svg'
    ]},
    author='Ibon',
    author_email='ibonescartin@gmail.com',
    keywords=['avatar', 'svg', 'python', 'face', 'custom'],
    url='https://github.com/ibonn/python_avatars',
    download_url='https://pypi.org/project/python_avatars/',
    project_urls={
        "Documentation": "https://readthedocs.org/projects/python_avatars/",
        "Source Code": "https://github.com/ibonn/python_avatars"
    },
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Topic :: Multimedia',
        'Topic :: Multimedia :: Graphics',
        'Topic :: Software Development :: Libraries :: Python Modules'
    ],
    include_package_data=True
)

install_requires = []

if __name__ == '__main__':
    setup(**setup_args, install_requires=install_requires)



from .base_enums import AvatarPart


class HatType(AvatarPart):
    """Hat types"""

    __install__ = True
    __enum_path__ = 'hat_types.py'
    __path__ = 'avatar_parts/top/hat'

    HAT = 'hat'
    HIJAB = 'hijab'
    TURBAN = 'turban'
    WINTER_HAT_1 = 'winter_hat_1'
    WINTER_HAT_2 = 'winter_hat_2'
    WINTER_HAT_3 = 'winter_hat_3'
    WINTER_HAT_4 = 'winter_hat_4'



from .base_enums import AvatarPart


class ClothingType(AvatarPart):
    """Clothing types"""

    __install__ = True
    __enum_path__ = 'clothing_types.py'
    __path__ = 'avatar_parts/clothes'

    NONE = ''
    BLAZER_SHIRT = 'blazer_shirt'
    BLAZER_SWEATER = 'blazer_sweater'
    COLLAR_SWEATER = 'collar_sweater'
    GRAPHIC_SHIRT = 'graphic_shirt'
    HOODIE = 'hoodie'
    OVERALL = 'overall'
    SHIRT_CREW_NECK = 'shirt_crew_neck'
    SHIRT_SCOOP_NECK = 'shirt_scoop_neck'
    SHIRT_V_NECK = 'shirt_v_neck'



from .base_enums import AvatarPart


class AccessoryType(AvatarPart):
    """Accessories (Glasses)"""

    __install__ = True
    __enum_path__ = 'accessory_types.py'
    __path__ = 'avatar_parts/accessories'

    NONE = ''
    EYEPATCH = 'eyepatch'
    KURT = 'kurt'
    PRESCRIPTION_1 = 'prescription_01'
    PRESCRIPTION_2 = 'prescription_02'
    ROUND = 'round'
    SUNGLASSES = 'sunglasses'
    WAYFARERS = 'wayfarers'
